@@ -21,17 +21,18 @@ from aioredis_cluster.command_info import (
     default_registry,
 )
 from aioredis_cluster.errors import RedisError, network_errors
-from aioredis_cluster.log import logger
 from aioredis_cluster.pooler import Pooler
 from aioredis_cluster.structs import Address, ClusterNode, ClusterSlot
 
-
 __all__ = (
     "ClusterManager",
     "create_cluster_state",
 )
 
 
+logger = logging.getLogger(__name__)
+
+
 SlotsResponse = List[List]
 
 
@@ -149,7 +150,7 @@ class ClusterManager:
 
         self._reload_count = 0
         self._reload_event = asyncio.Event()
-        self._loop = asyncio.get_event_loop()
+        self._loop = asyncio.get_running_loop()
         self._reloader_task = self._loop.create_task(self._state_reloader())
 
     @property
